
from enum import IntEnum

class MessageLength(IntEnum):
    UNDETERMINED = 0
    COMMAND = 13
    RESPONSE = 395

# remove stuff from here if it's only ever used once
# also split it into a few different, sensically grouped enums rather
# than encompassing everything here (I've done this now with CommandMessageType)
class CommandMessageConstants(IntEnum):
    BYTE_0 = 85                  # Byte 0 of command always fixed
    BYTE_2 = 12                  # Byte 2 of command always fixed
    TOGGLE = 128                 # Toggle on/off in byte 4
    BYTE_5_GRP_POS = 1           # In Position dec/inc command, byte 5 is fixed
    BYTE_4_GRP_POSDEC = 1        # Position decrement (5% down) in byte 4
    BYTE_4_GRP_POSINC = 2        # Position increment (5% up) in byte 4

# Go in byte 1 of command messages
class CommandMessageType(IntEnum):
    REQUEST_STATE = 1
    ZONE_CONROL = 129
    AC_CONTROL = 134

# Go in byte 4 of CommandMessageType.AC_CONTROL messages 
class ACControlCommands:
    SET_MODE = 129
    SET_FAN_SPEED = 130
    TEMP_DEC = 147
    TEMP_INC = 163

class ResponseMessageConstants(IntEnum):
    LONG_STRING_LENGTH = 16
    SHORT_STRING_LENGTH = 8

# Thanks to home-assistant.io user Radebe2k
class ResponseMessageOffsets(IntEnum):
    # Header is 4 bytes
    HEADER = 0

    # There are 16 zones
    # zone names are 8 bytes (ResponseMessageConstants.SHORT_STRING_LENGTH)
    ZONE_NAMES_START = 100
    # zone statuses are 1 byte each
    ZONE_STATUSES_START = 228
    # Zone strengths are 1 byte each
    ZONE_STRENGTHS_START = 276
    
    # System name is 16 bytes (ResponseMessageConstants.LONG_STRING_LENGTH)
    SYSTEM_NAME = 324

    # perhaps should just store the first offset and iterate from 1st to 2nd
    AC1_STATUS = 354
    AC2_STATUS = 355
    AC1_MODE = 358
    AC2_MODE = 359
    AC1_FAN_SPEED = 360
    AC2_FAN_SPEED = 361
    AC1_SET_TEMP = 362
    AC2_SET_TEMP = 363
    AC1_AMBIENT_TEMP = 364
    AC2_AMBIENT_TEMP = 365
    AC1_MANUFACTURER = 356
    AC2_MANUFACTURER = 357
    # AC names are 8 bytes (ResponseMessageConstants.SHORT_STRING_LENGTH)
    AC1_NAME_START = 370
    AC2_NAME_START = 378
    HASH = 394



from __future__ import annotations
from typing import TYPE_CHECKING
from airtouch2.protocol.enums import ACFanSpeed, ACManufacturer, ACMode
from airtouch2.protocol.messages import ChangeSetTemperature, ResponseMessage, SetFanSpeed, SetMode, ToggleAC
if TYPE_CHECKING:
    from airtouch2.AT2Client import AT2Client

class AT2Aircon:
    def __init__(self, number: int, client: AT2Client, response_message: ResponseMessage=None):
        self.number = number
        self._client = client
        if response_message:
            self.update(response_message)
        else:
            self.system_name = "UNKNOWN"
            self.name = "UNKNOWN"
            self.manufacturer = ACManufacturer.NONE
            self.on = False
            self.mode = -1
            self.fan_speed = -1
            self.ambient_temp = -1
            self.set_temp = -1
            self.zones = []

    def update(self, response_message: ResponseMessage) -> None:
        self.system_name = response_message.system_name
        self.on = response_message.ac_active[self.number]
        self.status = response_message.ac_status[self.number]
        self.mode = response_message.ac_mode[self.number]
        self.fan_speed = response_message.ac_fan_speed[self.number]
        self.ambient_temp = response_message.ac_ambient_temp[self.number]
        self.set_temp = response_message.ac_set_temp[self.number]
        self.manufacturer = response_message.ac_manufacturer[self.number]
        self.name = response_message.ac_name[self.number]

    def inc_dec_set_temp(self, inc: bool):
        self._client.send_command(ChangeSetTemperature(self.number, inc))

    def set_set_temp(self, new_temp: int):
        temp_diff = new_temp - self.set_temp
        #print(f"Temp diff: {temp_diff}")
        inc = temp_diff > 0
        for i in range(abs(temp_diff)):
            self.inc_dec_set_temp(inc)

    def _turn_on_off(self, on: bool):
        if self.on != on:
            self._client.send_command(ToggleAC(self.number))
    
    def turn_off(self):
        self._turn_on_off(False)
    
    def turn_on(self):
        self._turn_on_off(True)

    def set_fan_speed(self, fan_speed: ACFanSpeed):
        self._client.send_command(SetFanSpeed(self.number, fan_speed))

    def set_mode(self, mode: ACMode):
        self._client.send_command(SetMode(self.number, mode))

    def __str__(self):
        return f"""
        System Name:\t{self.system_name}
        AC Name:\t{self.name}
        On:\t\t{self.on}
        Status:\t\t{self.status}
        Mode:\t\t{self.mode}
        Fan Speed:\t{self.fan_speed}
        Ambient Temp:\t{self.ambient_temp}
        Set Temp:\t{self.set_temp}
        Manufacturer:\t{self.manufacturer}
        """
        


from __future__ import annotations
from abc import ABC
from airtouch2.protocol.constants import ACControlCommands, CommandMessageConstants, CommandMessageType, MessageLength, ResponseMessageConstants, ResponseMessageOffsets
from airtouch2.protocol.enums import ACFanSpeed, ACManufacturer, ACMode
from sys import maxsize as MAX_INT

class Message(ABC):
    """ Message base class from which all airtouch 2 communications messages are derived"""
    length: MessageLength = MessageLength.UNDETERMINED

    def add_checksum(self, serial_msg: bytearray) -> bytearray:
        serial_msg[self.length-1] = self.checksum(serial_msg)
        return serial_msg
    
    @staticmethod
    def checksum(serial_msg: bytearray) -> int:
        sum: int = 0
        for b in serial_msg:
            sum += b
        sum = (sum % 256)
        return sum

class CommandMessage(Message):
    """ Command message base class from which all airtouch2 command messages are derived"""
    length: MessageLength = MessageLength.COMMAND
    
    # command messages are only sent so only need to be serialized
    def serialize(self, prefilled_msg: bytearray) -> bytearray:
        prefilled_msg[0] = CommandMessageConstants.BYTE_0
        prefilled_msg[2] = CommandMessageConstants.BYTE_2
        return self.add_checksum(prefilled_msg)

class RequestState(CommandMessage):
    """ Command to request the state of the airtouch 2 system"""
    
    def serialize(self) -> bytearray:
        serial_msg: bytearray = bytearray(self.length)
        serial_msg[1] = CommandMessageType.REQUEST_STATE
        return super().serialize(serial_msg)

class ACControlCommand(CommandMessage):
    """Base class from which all AC control messages are derived"""
    def __init__(self, target_ac_number: int):
        super().__init__()
        self.target_ac = target_ac_number

    def serialize(self, prefilled_msg: bytearray) -> bytearray:
        prefilled_msg[1] = CommandMessageType.AC_CONTROL
        prefilled_msg[3] = self.target_ac
        return super().serialize(prefilled_msg)

class ChangeSetTemperature(ACControlCommand):
    """Command to increment or decrement the AC set point by 1 degree"""
    def __init__(self, target_ac_number: int, inc: bool):
        super().__init__(target_ac_number)
        self.inc = inc

    def serialize(self) -> bytearray:
        serial_msg: bytearray = bytearray(self.length)
        serial_msg[4] = ACControlCommands.TEMP_INC if self.inc else ACControlCommands.TEMP_DEC
        return super().serialize(serial_msg)

class ToggleAC(ACControlCommand):
    """Command to toggle an AC on or off"""
    
    def serialize(self) -> bytearray:
        serial_msg: bytearray = bytearray(self.length)
        serial_msg[4] = CommandMessageConstants.TOGGLE
        return super().serialize(serial_msg)

# needs investigation
class SetFanSpeed(ACControlCommand):
    """Command to set the AC fan speed to one of those in ACFanSpeed"""
    def __init__(self, target_ac_number: int, fan_speed: ACFanSpeed):
        super().__init__(target_ac_number)
        self.fan_speed: ACFanSpeed = fan_speed

    def serialize(self) -> bytearray:
        serial_msg: bytearray = bytearray(self.length)
        serial_msg[4] = ACControlCommands.SET_FAN_SPEED
        serial_msg[5] = self.fan_speed
        return super().serialize(serial_msg)

class SetMode(ACControlCommand):
    """Command to set the AC mode to one of those in ACMode"""
    def __init__(self, target_ac_number: int, mode: ACMode):
        super().__init__(target_ac_number)
        self.mode: ACMode = mode
    
    def serialize(self) -> bytearray:
        serial_msg: bytearray = bytearray(self.length)
        serial_msg[4] = ACControlCommands.SET_MODE
        serial_msg[5] = self.mode
        return super().serialize(serial_msg)

class ResponseMessage(Message):
    """ The airtouch 2 response message (there is only one) that contains all the information about the current state of the system"""
    
    length: MessageLength = MessageLength.RESPONSE

    # only getting fundamental things for AC1 for now
    def __init__(self, raw_response: bytes):
        super().__init__()
        # seems to be x000_1111 - x is on/off. This is different to what Luke describes.
        status = raw_response[ResponseMessageOffsets.AC1_STATUS]
        self.ac_active = [(status & 0x80 > 0)]
        self.ac_status = [(status & 0x7F)]
        # simply 0-4, see ACMode enum
        self.ac_mode = [ACMode(raw_response[ResponseMessageOffsets.AC1_MODE])]
        # 0100_0xxx - least significant byte is 0 to 4 - see ACFanSpeed enum
        self.ac_fan_speed = [ACFanSpeed(raw_response[ResponseMessageOffsets.AC1_FAN_SPEED] & 0x0F)]
        self.ac_set_temp = [raw_response[ResponseMessageOffsets.AC1_SET_TEMP]]
        self.ac_ambient_temp = [raw_response[ResponseMessageOffsets.AC1_AMBIENT_TEMP]]
        self.ac_manufacturer = [ACManufacturer(raw_response[ResponseMessageOffsets.AC1_MANUFACTURER])]
        self.ac_name = [raw_response[ResponseMessageOffsets.AC1_NAME_START:ResponseMessageOffsets.AC1_NAME_START+ResponseMessageConstants.SHORT_STRING_LENGTH].decode()]
        #self.zones = {}
        self.system_name = raw_response[ResponseMessageOffsets.SYSTEM_NAME:ResponseMessageOffsets.SYSTEM_NAME+ResponseMessageConstants.LONG_STRING_LENGTH].decode()



import errno
import socket
import logging

from threading import Thread, Event
from queue import Empty, Queue
from time import sleep
from typing import Callable

from airtouch2.AT2Aircon import AT2Aircon

from airtouch2.protocol.constants import MessageLength
from airtouch2.protocol.messages import CommandMessage, RequestState, ResponseMessage

SocketClosedErrors = (errno.ECONNABORTED,  errno.ECONNRESET, errno.ENOTCONN, errno.ESHUTDOWN, errno.ECONNREFUSED)

_LOGGER = logging.getLogger(__name__)

class AT2Client:

    def __init__(self, host: str):
        self._host_ip: str = host
        self._host_port: int = 8899
        self._sock: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.settimeout(None)
        self._stop_threads: bool = True
        self._cmd_queue: Queue[CommandMessage] = Queue()
        self._last_response: ResponseMessage = None
        self._new_response: Event =  Event()
        self._new_response_or_command: Event = Event()
        self.aircons: list[AT2Aircon] = []
        self.system_name = "UNKNOWN"
        self._threads: list[Thread] = []
        self._active: bool = False
        self._callbacks: list[Callable] = []
        self._socket_broken = False
        self._data_updated = Event()

    def __del__(self):
        self.stop()
    
    def start(self) -> bool:
        if not self._connect():
            return False
        self._threads = [Thread(target=self._handle_incoming), Thread(target=self._main_loop)]
        self._stop_threads = False
        for t in self._threads:
            t.start()
        self._active = True
        self.update_state()
        return True

    def _connect(self) -> bool:
        _LOGGER.debug(f'Connecting to {self._host_ip} on port {self._host_port}')
        try:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._sock.settimeout(None)
            self._sock.connect((self._host_ip, self._host_port))
        except OSError as e:
            _LOGGER.warning(f"Could not connect to host {self._host_ip}")
            if isinstance(e, socket.gaierror):
                pass
            elif e.errno not in (errno.EHOSTUNREACH, errno.ECONNREFUSED,  errno.ETIMEDOUT):
                raise e
            self._sock.close()
            return False
        _LOGGER.debug("Connected to airtouch2 server")
        return True

    def _stop_common(self):
        if self._active:
            self._stop_threads = True
            _LOGGER.debug("Closing socket...")
            try:
                self._sock.shutdown(socket.SHUT_RDWR)
            except OSError as e:
                if e.errno not in SocketClosedErrors:
                    raise e
                _LOGGER.debug(f"Socket shutdown() call failed:\n  OSError: [Errno {e.errno}] {e.strerror}")
            self._sock.close()
            self._data_updated.clear()
            self._new_response_or_command.set()

    def _reset(self):
        self._new_response_or_command.clear()
        self._new_response.clear()
        self._data_updated.clear()
        self.aircons.clear()
        self._threads.clear()
        self._socket_broken = False
        self._active = False
        self._last_response = None
        self._stop_threads = True
        self.system_name = "UNKNOWN"
        while not self._cmd_queue.empty():
            try:
                self._cmd_queue.get_nowait()
            except Empty:
                pass

    def _stop_and_join_main_thread(self):
        self._stop_common()
        _LOGGER.debug("Joining main thread")
        self._threads[1].join()
        self._reset()

    def stop(self):
        self._stop_common()
        _LOGGER.debug("Joining threads...")
        for t in self._threads:
            t.join()
        self._reset()
        _LOGGER.debug("Shutdown successful")


    def add_callback(self, func: Callable):
        self._callbacks.append(func)
        def remove_callback() -> None:
            if func in self._callbacks:
                self._callbacks.remove(func)

        return remove_callback

    def send_command(self, command: CommandMessage) -> None:
        self._cmd_queue.put(command)
        self._new_response_or_command.set()

    def update_state(self):
        self._data_updated.clear()
        self.send_command(RequestState())
        self._data_updated.wait()

    def _await_response(self) -> bytes:
        # taken (almost) straight from https://docs.python.org/3/howto/sockets.html
        chunks = []
        bytes_recd = 0
        while bytes_recd < MessageLength.RESPONSE:
            try:
                chunk = self._sock.recv(MessageLength.RESPONSE - bytes_recd) if not self._stop_threads else b''
            except OSError as e:
                if e.errno not in SocketClosedErrors:
                    raise e
                chunk=b''
            if chunk == b'':
                _LOGGER.debug("Socket closed/broken")
                break
            chunks.append(chunk)
            bytes_recd = bytes_recd + len(chunk)
        return b''.join(chunks)

    def _handle_incoming(self) -> None:
        while not self._stop_threads:
            resp = self._await_response()
            if len(resp) == 0:
                self._socket_broken = True
                break
            if len(resp) != MessageLength.RESPONSE:
                _LOGGER.warning("Received invalid response, ignoring")
                continue
            _LOGGER.debug("handle_incoming received new response")
            self._last_response = ResponseMessage(resp)
            self._new_response.set()
            self._new_response_or_command.set()

        if not self._stop_threads and self._socket_broken:
            _LOGGER.debug("Socket connection was broken, trying to recover...")
            self._stop_and_join_main_thread()
            _LOGGER.debug("Restarting client...")
            retries = 0
            while not self.start():
                sleep(0.001 * (10**retries) if retries < 4 else 10)
                retries+=1
                if not retries % 50:
                    _LOGGER.debug("Server is not responding, will continue trying to reconnect every 10s")

    def _process_last_response(self):
        # only process if we have a new one
        if self._last_response:
            resp = self._last_response
            self._last_response = None
            self._new_response.clear()
            self._new_response_or_command.clear()
            # check message if there are 1 or 2 aircons - not sure how to do this yet
            # do the stuff, update aircons, zones etc...
            if not self.aircons:
                self.aircons.append(AT2Aircon(0, self, resp))
                _LOGGER.debug(self.aircons[0])
            else:
                for aircon in self.aircons:
                    aircon.update(resp)
                    _LOGGER.debug(aircon)
            self._data_updated.set()
            for func in self._callbacks:
                _LOGGER.debug("Triggering callback")
                func()

    def _main_loop(self) -> None:
        """Main loop"""
        while not self._stop_threads:
            # wait for either a server message to be emitted or a command to be put on the queue
            _LOGGER.debug("Waiting for new data to receive or command to send")
            self._new_response_or_command.wait()
            _LOGGER.debug("Main loop waiter Event triggered")
            # process a message if that's what triggered the event
            self._process_last_response()
            
            # process commands if that's what triggered the event
            while self._cmd_queue.qsize() > 0 and not self._socket_broken:
                _LOGGER.debug("There are commands in the queue")
                try:
                    cmd = self._cmd_queue.get(block=False)
                except Empty:
                    _LOGGER.debug("Command queue was empty")
                    break
                self._new_response.clear()
                _LOGGER.debug(f"Sending {cmd.__class__.__name__}")
                if not self._socket_broken:
                    self._sock.sendall(cmd.serialize())
                    self._new_response.wait()
                    self._process_last_response()



from enum import IntEnum

class ACMode(IntEnum):
    AUTO = 0
    HEAT = 1
    DRY = 2
    FAN = 3
    COOL = 4

    def __str__(self):
        return self._name_

class ACFanSpeed(IntEnum):
    AUTO = 0
    QUIET = 1
    LOW = 2
    MEDIUM = 3
    HIGH = 4

    def __str__(self):
        return self._name_

#Daikin and Panasonic from Luke's images
class ACManufacturer(IntEnum):
    NONE = 0
    DAIKIN = 1
    FUJITSU = 2
    PANASONIC = 7
    # NO IDEA ABOUT THESE, just going by the order listed in the installation manual
    HITACHI = 3
    LG = 4
    MITSUBISHI_ELECTRIC = 5
    # from the manual I would expect panasonic to be 6, not 7
    #PANASONIC = 6
    # added 1 to these as they come after panasonic, which is 7 now, not 6
    SAMSUNG = 8
    TOSHIBA = 9
    MITSUBISHI_HEAVY_IND = 10
        
    def __str__(self):
        return self._name_